import re

# Code extracted from:
# https://github.com/django/django/blob/stable/1.3.x/django/core/validators.py#L45
url_regex = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)


def is_valid_url(value: str) -> bool:
    """
    Check if the provided string is a valid URL.

    Args:
        value (str): The string to be validated.

    Returns:
        bool: True if the string is a valid URL, False otherwise.
    """

    return re.match(url_regex, value) is not None



from typing import Callable
from functools import wraps
from textual.widgets.data_table import RowDoesNotExist, CellDoesNotExist


def fetch_guard(func: Callable) -> Callable:
    """
    Decorator to prevent multiple fetch request at the same time and
    avoid executing certain actions while a fetch is in progress.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if self.is_fetching:
            self.notify(
                "a data refresh is in progress... please wait until it finishes",
                severity="warning",
            )
            return

        return await func(self, *args, **kwargs)

    return wrapper


def rollback_session(
    error_message: str = "",
    severity: str = "error",
    callback: Callable | None = None,
) -> Callable:
    """
    Decorator to handle exceptions and perform a rollback if needed. It also
    notifies the user with an error message and, if specified, executes a callback
    function at the end.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            try:
                return await func(self, *args, **kwargs)
            except (RowDoesNotExist, CellDoesNotExist):
                pass
            except Exception as e:
                self.session.rollback()
                message = (
                    f"{error_message}: {e}"
                    if error_message
                    else f"something went wrong: {e}"
                )
                self.notify(message, severity=severity)
            finally:
                if callback:
                    callback(self)

        return wrapper

    return decorator



import shutil
import click
from importlib.metadata import version
from pathlib import Path
from typing import Type, Tuple
from pydantic import BaseModel, Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

APP_NAME = "lazyfeed"
APP_DIR = Path(click.get_app_dir(app_name=APP_NAME))
DB_URL: Path = APP_DIR / f"{APP_NAME}.db"
CONFIG_FILE_PATH = APP_DIR / "config.toml"
TEMPLATE_FILE_PATH = Path(__file__).parent / "config_template.toml"


class ClientSettings(BaseModel):
    timeout: int = 300
    connect_timeout: int = 10
    headers: dict = {}


class Settings(BaseSettings):
    name: str = APP_NAME
    version: str = version(APP_NAME)

    http_client: ClientSettings = Field(default_factory=ClientSettings)

    theme: str = "dracula"

    auto_read: bool = False
    auto_load: bool = False
    confirm_before_read: bool = True
    sort_by: str = "published_at"
    sort_order: str = "ascending"

    model_config = SettingsConfigDict(
        env_nested_delimiter="__",
        toml_file=f"{APP_DIR / 'config.toml'}",
        validate_default=True,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        APP_DIR.mkdir(parents=True, exist_ok=True)

        if not CONFIG_FILE_PATH.exists():
            shutil.copy(TEMPLATE_FILE_PATH, CONFIG_FILE_PATH)

        return (
            env_settings,
            TomlConfigSettingsSource(settings_cls),
        )


if __name__ == "__main__":
    settings = Settings()
    print(settings.model_dump())



from textual.message import Message


class AddFeed(Message):
    """Message to add a new RSS feed."""

    pass


class EditFeed(Message):
    """Message to edit an existing RSS feed."""

    def __init__(self, id: int) -> None:
        self.id = id
        super().__init__()


class DeleteFeed(Message):
    """Message to delete a specified RSS feed."""

    def __init__(self, id: int) -> None:
        self.id = id
        super().__init__()


class FilterByFeed(Message):
    """Message to filter by the specified RSS feed."""

    def __init__(self, id: int) -> None:
        self.id = id
        super().__init__()


class MarkAsRead(Message):
    """Message to mark an item as 'read'."""

    def __init__(self, item_id: int) -> None:
        self.item_id = item_id
        super().__init__()


class MarkAllAsRead(Message):
    """Message to mark all items as 'read'."""

    pass


class Open(Message):
    """Message to open item's content."""

    def __init__(self, item_id: int) -> None:
        self.item_id = item_id
        super().__init__()


class OpenInBrowser(Message):
    """Message to open an item in the browser."""

    def __init__(self, item_id: int) -> None:
        self.item_id = item_id
        super().__init__()


class SaveForLater(Message):
    """Message to save an item for later."""

    def __init__(self, item_id: int) -> None:
        self.item_id = item_id
        super().__init__()


class ShowPending(Message):
    """Message to list all items."""

    pass


class ShowAll(Message):
    """Message to list all pending items."""

    pass


class ShowSavedForLater(Message):
    """Message to list all saved for later items."""

    pass


class ShowToday(Message):
    """Message to list all items published at today's date."""

    pass



from textual import on
from textual.app import ComposeResult
from textual.containers import Container, VerticalScroll
from textual.validation import Function
from textual.widgets import Button, Input, Label
from textual.screen import ModalScreen
from lazyfeed.widgets.validators import is_valid_url


class EditFeedModal(ModalScreen[dict | None]):
    """
    Modal screen for editing an existing RSS feed.
    """

    BINDINGS = [
        ("escape,q", "dismiss", "dismiss"),
    ]

    def __init__(self, url: str, title: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.url = url
        self.title = title

    def compose(self) -> ComposeResult:
        with VerticalScroll(classes="modal-body") as container:
            container.border_title = "edit feed"

            yield Container(
                Container(
                    Label("feed title"),
                    Input(
                        placeholder="title (optional)",
                        classes="input--feed-title",
                        value=self.title,
                    ),
                ),
                Container(
                    Label("url"),
                    Input(
                        placeholder="url",
                        validate_on=["changed", "submitted"],
                        validators=[
                            Function(is_valid_url, "invalid url."),
                        ],
                        classes="input--feed-url",
                        value=self.url,
                    ),
                ),
                classes="inputs",
            )
            yield Button(label="update", variant="success")

    def on_mount(self) -> None:
        self.button = self.query_one(Button)

    def action_dismiss_overlay(self) -> None:
        self.dismiss(None)

    @on(Input.Changed, ".input--feed-url")
    def enable_button(self, event: Input.Changed) -> None:
        if not event.value or not event.validation_result.is_valid:
            self.button.disabled = True
            return

        self.button.disabled = False

    @on(Button.Pressed)
    def submit_form(self) -> None:
        self.dismiss(
            {
                "title": self.query_one(".input--feed-title").value,
                "url": self.query_one(".input--feed-url").value,
            }
        )

    @on(Input.Submitted, ".input--feed-url")
    async def edit_feed(self, event: Input.Submitted) -> None:
        if not event.validation_result.is_valid:
            return

        self.dismiss(
            {
                "title": self.query_one(".input--feed-title").value,
                "url": event.value,
            }
        )



import aiohttp
from contextlib import asynccontextmanager
from lazyfeed.settings import Settings


@asynccontextmanager
async def http_client_session(settings: Settings):
    """
    Asynchronous context manager for creating an HTTP client session using
    aiohttp.ClientSession. It configures the session with specified timeouts
    and headers from the provided settings, and the session is automatically
    closed upon exiting the context.
    """

    client_timeout = aiohttp.ClientTimeout(
        total=settings.http_client.timeout,
        connect=settings.http_client.connect_timeout,
    )

    async with aiohttp.ClientSession(
        timeout=client_timeout,
        headers=settings.http_client.headers,
    ) as session:
        yield session



from lazyfeed.widgets.modals.add_feed_modal import AddFeedModal
from lazyfeed.widgets.modals.confirm_action_modal import ConfirmActionModal
from lazyfeed.widgets.modals.edit_feed_modal import EditFeedModal
from lazyfeed.widgets.modals.help_modal import HelpModal

__all__ = [
    "AddFeedModal",
    "ConfirmActionModal",
    "EditFeedModal",
    "HelpModal",
]



from lazyfeed.widgets.custom_header import CustomHeader
from lazyfeed.widgets.item_table import ItemTable
from lazyfeed.widgets.item_screen import ItemScreen
from lazyfeed.widgets.rss_feed_tree import RSSFeedTree

__all__ = [
    "CustomHeader",
    "ItemTable",
    "ItemScreen",
    "RSSFeedTree",
]



from textual import on
from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Button, Static
from textual.screen import ModalScreen


class ConfirmActionModal(ModalScreen[bool]):
    """
    Modal screen with a prompt for confirmation of an action.
    """

    BINDINGS = [
        ("escape,q", "dismiss", "dismiss"),
        ("enter,y", "confirm", "confirm"),
    ]

    def __init__(
        self, border_title: str, message: str, action_name: str, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.border_title = border_title
        self.message = message
        self.action_name = action_name

    def compose(self) -> ComposeResult:
        with VerticalScroll(classes="modal-body modal-body--error") as container:
            container.border_title = self.border_title

            yield Static(self.message)
            yield Button(label=self.action_name, variant="error")

    def on_mount(self) -> None:
        self.query_one(Button).focus()

    def action_dismiss_overlay(self) -> None:
        self.dismiss(False)

    @on(Button.Pressed)
    def action_confirm(self) -> None:
        self.dismiss(True)



from rich.text import Text
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import DataTable, Label, Markdown
from lazyfeed.widgets.helpable import Helpable


class HelpModal(ModalScreen[None]):
    """
    Modal screen to display help information for a specific widget.
    """

    BINDINGS = [
        ("escape,q", "dismiss", "dismiss"),
    ]

    def __init__(self, widget: Widget) -> None:
        super().__init__()
        self.widget = widget

    def compose(self) -> ComposeResult:
        with VerticalScroll(classes="modal-body") as container:
            widget = self.widget
            if isinstance(widget, Helpable):
                help = widget.help
                help_title = help.title
                help_description = help.description

                if help_title:
                    container.border_title = help_title
                else:
                    container.border_title = "help"

                if help_description:
                    help_description = help_title.strip()
                    with VerticalScroll(classes="help-description"):
                        yield Markdown(help_description)

            bindings = widget._bindings
            keys: list[tuple[str, list[Binding]]] = list(
                bindings.key_to_bindings.items(),
            )

            if keys:
                yield Label("all keybindings", classes="help-table__label")

                table = DataTable(cursor_type="row", zebra_stripes=True)
                table.add_columns("key", "description")
                for _, bindings in keys:
                    table.add_row(
                        Text(
                            ", ".join(
                                binding.key_display
                                if binding.key_display
                                else self.app.get_key_display(binding)
                                for binding in bindings
                            ),
                            style="bold",
                            no_wrap=True,
                            end="",
                        ),
                        bindings[0].description.lower(),
                    )

                yield table



from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import Screen
from textual.widgets import MarkdownViewer, Footer
from lazyfeed.messages import OpenInBrowser, SaveForLater
from lazyfeed.models import Item


class ItemScreen(Screen):
    """
    Screen for displaying the content of a single item in markdown format.
    """

    BINDINGS = [
        Binding("R", "none", "none", show=False, priority=True),
        Binding("ctrl+c,escape,q,o", "app.pop_screen", "go back", priority=True),
        Binding("s", "save_for_later", "save", priority=True),
        Binding("O", "open_in_browser", "open in browser", priority=True),
    ]

    def __init__(self, item: Item):
        super().__init__()
        self.item = item

    def compose(self) -> ComposeResult:
        yield MarkdownViewer(self.item.content, show_table_of_contents=False)
        yield Footer()

    def on_mount(self) -> None:
        self.md_viewer = self.query_one(MarkdownViewer)
        self.md_viewer.border_title = "article"

    def action_open_in_browser(self) -> None:
        self.post_message(OpenInBrowser(self.item.id))

    def action_save_for_later(self) -> None:
        self.post_message(SaveForLater(self.item.id))

    def action_none(self) -> None:
        pass



from datetime import datetime
import aiohttp
import feedparser
from selectolax.parser import HTMLParser
from markdownify import markdownify as md
from lazyfeed.models import Feed, Item


def clean_html(html: str) -> str | None:
    """
    Removes unwanted content from the given HTML.

    Args:
        html (str): The HTML content to clean.

    Returns:
        str | None: The cleaned HTML content, or None in the case of error.
    """

    tree = HTMLParser(html)
    tags = [
        "canvas",
        "footer",
        "head",
        "header",
        "iframe",
        "nav",
        "script",
        "style",
        "noscript",
    ]
    tree.strip_tags(tags)
    return tree.html


async def fetch_feed(
    client: aiohttp.ClientSession,
    url: str,
    title: str | None = None,
) -> Feed:
    """
    Fetch and parse an RSS feed from the specified URL.

    Args:
        client (aiohttp.ClientSession): The HTTP client session to use for the request.
        url (str): The URL of the feed to be fetched.
        title (str | None): Optional title for the feed.

    Returns:
        Feed: Feed object.

    Raises:
        RuntimeError: If the feed cannot be fetched or is badly formatted.
    """

    try:
        resp = await client.get(url)
        resp.raise_for_status()
    except aiohttp.ClientError as e:
        raise RuntimeError(f'failed to fetch feed from "{url}": {e}')

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise RuntimeError(f"feed is badly formatted: {d.bozo_exception}")

    metadata = d["channel"]
    feed = Feed(
        url=url,
        title=title or metadata.get("title"),
        site=metadata.get("link"),
        description=metadata.get("description", ""),
    )

    return feed


async def fetch_entries(
    client: aiohttp.ClientSession,
    url: str,
    etag: str = "",
) -> tuple[list[dict], str]:
    """
    Fetch entries from the specified RSS feed URL.

    Args:
        client (aiohttp.ClientSession): The HTTP client session to use for the request.
        url (str): The URL of the feed to fetch entries from.
        etag (str): Optional ETag header to check if the feed has been updates since the last time.

    Returns:
        tuple[list[dict], str]: A tuple containing a list of entries and the ETag from the response.

    Raises:
        RuntimeError: If the feed cannot be fetched or is badly formatted.
    """

    headers = {}
    if etag:
        headers["If-None-Match"] = etag

    try:
        resp = await client.get(url, headers=headers)
        resp.raise_for_status()
    except aiohttp.ClientError as e:
        raise RuntimeError(f'failed to fetch items from "{url}": {e}')

    if resp.status == 304:
        return [], etag

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise RuntimeError(f"feed is badly formatted: {d.bozo_exception}")

    return d.get("entries", []), resp.headers.get("Etag", "")


async def fetch_content(
    client: aiohttp.ClientSession,
    entry_data: dict,
    feed_id: int,
) -> Item | None:
    """
    Fetch and parse the content of a specific entry.

    Args:
        client (aiohttp.ClientSession): The HTTP client session to use for the request.
        entry_data (dict): The data of the entry to fetch content for.
        feed_id (int): The ID of the feed associated with the entry.

    Returns:
        Item | None: An Item object containing the entry's content, or None if the entry is invalid or something went wrong while parsing.

    Raises:
        RuntimeError: If the content cannot be fetched.
    """

    url = entry_data.get("link")
    title = entry_data.get("title", "")
    author = entry_data.get("author", "")
    description = entry_data.get("description", "")
    published_parsed = entry_data.get("published_parsed")

    assert url

    try:
        resp = await client.get(url)
        resp.raise_for_status()
    except aiohttp.ClientError as e:
        raise RuntimeError(f'failed to fetch contents from "{url}": {e}')

    raw_content = await resp.text()
    md_content = md(clean_html(raw_content))

    published_at = None
    if published_parsed:
        published_at = datetime(*published_parsed[:6])

    return Item(
        title=title,
        url=url,
        author=author,
        description=description,
        raw_content=raw_content,
        content=md_content,
        feed_id=feed_id,
        published_at=published_at,
    )



import asyncio
import sys
from sqlalchemy import select
from sqlalchemy.orm import Session
from lazyfeed.app import LazyFeedApp
from lazyfeed.feeds import fetch_feed
from lazyfeed.http_client import http_client_session
from lazyfeed.models import Feed
from lazyfeed.settings import Settings
from lazyfeed.utils import export_opml, import_opml, console


async def fetch_new_feeds(
    settings: Settings,
    session: Session,
    feeds: set[str],
) -> None:
    """
    Fetch and store new RSS feeds.
    """

    async with http_client_session(settings) as client_session:
        tasks = [fetch_feed(client_session, feed) for feed in feeds]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        for result in results:
            if isinstance(result, Exception):
                console.print(f"❌ something went wrong fetching feed: {result}")
                continue

            try:
                session.add(result)
                session.commit()
                console.print(f'✅ added "{result.url}"')
            except Exception as e:
                session.rollback()
                console.print(
                    f"❌ something went wrong while saving feeds to the database: {e}"
                )


def main():
    settings = Settings()
    app = LazyFeedApp(settings)
    session = app.session

    if not sys.stdin.isatty():
        with console.status(
            "[green]importing feeds from file... please, wait a moment",
            spinner="earth",
        ) as status:
            opml_content = sys.stdin.read()
            feeds_in_file = import_opml(opml_content)

            console.print("✅ file read correctly")

            stmt = select(Feed.url)
            results = session.execute(stmt).scalars().all()
            new_feeds = {feed for feed in feeds_in_file if feed not in results}
            if not new_feeds:
                console.print("✅ all feeds had been already added")
                return

            status.update(f"[green]fetching {len(new_feeds)} new feeds...[/]")
            asyncio.run(fetch_new_feeds(settings, session, new_feeds))
            return

    if not sys.stdout.isatty():
        stmt = select(Feed)
        results = session.execute(stmt).scalars().all()
        output = export_opml(results)
        sys.stdout.write(output)
        return

    app.run()


if __name__ == "__main__":
    main()



import io
import xml.etree.ElementTree as ET
from rich.console import Console
from lazyfeed.models import Feed
from lazyfeed.settings import APP_NAME

console = Console(emoji=True)


def export_opml(feeds: list[Feed]):
    """
    Export a list of RSS feeds into an OPML formatted string.
    """

    opml = ET.Element("opml", version="1.0")

    head = ET.SubElement(opml, "head")
    title = ET.SubElement(head, "title")
    title.text = f"RSS feeds from {APP_NAME}"
    body = ET.SubElement(opml, "body")

    for feed in feeds:
        ET.SubElement(
            body,
            "outline",
            text=feed.title,
            type="rss",
            xmlUrl=feed.url,
        )

    tree = ET.ElementTree(opml)
    output_buffer = io.BytesIO()

    tree.write(output_buffer, encoding="utf-8", xml_declaration=True)

    opml_output = output_buffer.getvalue().decode("utf-8")
    output_buffer.close()

    return opml_output


def import_opml(input: str) -> list[str]:
    """
    Import RSS feeds from an OPML formatted string.
    """

    feeds = []

    root = ET.fromstring(input)
    for outline in root.findall(".//outline"):
        xml_url = outline.get("xmlUrl")
        if xml_url:
            feeds.append(xml_url)

    return feeds



from datetime import datetime
from typing import List
from sqlalchemy import ForeignKey, Boolean, Text, func
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    mapped_column,
    relationship,
)


class Base(DeclarativeBase):
    pass


class Feed(Base):
    __tablename__ = "feed"

    id: Mapped[int] = mapped_column(primary_key=True)
    url: Mapped[str] = mapped_column(unique=True)
    site: Mapped[str] = mapped_column(nullable=True)
    title: Mapped[str]
    description: Mapped[str] = mapped_column(nullable=True)

    items: Mapped[List["Item"]] = relationship(
        back_populates="feed",
        cascade="all, delete",
    )

    etag: Mapped[str] = mapped_column(nullable=True)
    created_at: Mapped[datetime] = mapped_column(default=func.now())
    last_updated_at: Mapped[datetime] = mapped_column(
        default=func.now(),
        onupdate=func.now(),
    )

    def __repr__(self) -> str:
        return f"<Feed(id={self.id!r}, url={self.url!r}, site={self.site!r}, title={self.title!r}, created_at={self.created_at!r}, last_updated_at={self.last_updated_at!r})>"


class Item(Base):
    __tablename__ = "item"

    id: Mapped[int] = mapped_column(primary_key=True)
    title: Mapped[str] = mapped_column(nullable=True)
    url: Mapped[str] = mapped_column(unique=True)
    author: Mapped[str] = mapped_column(nullable=True)
    description: Mapped[str] = mapped_column(nullable=True)

    is_read: Mapped[bool] = mapped_column(Boolean(), default=False)
    is_saved: Mapped[bool] = mapped_column(Boolean(), default=False)

    feed_id: Mapped[int] = mapped_column(ForeignKey("feed.id"))
    feed: Mapped[Feed] = relationship(back_populates="items")

    raw_content: Mapped[str] = mapped_column(Text(), nullable=True)
    content: Mapped[str] = mapped_column(Text(), nullable=True)

    published_at: Mapped[datetime] = mapped_column(default=func.now())
    last_updated_at: Mapped[datetime] = mapped_column(
        default=func.now(),
        onupdate=func.now(),
    )

    def __repr__(self) -> str:
        return f"<Item(id={self.id!r}, url={self.url!r}, author={self.author!r}, title={self.title!r}, is_read={self.is_read!r}, is_favorite={self.is_favorite!r}, is_saved={self.is_saved!r}, published_at={self.published_at!r}, last_updated_at={self.last_updated_at!r})>"



import asyncio
from datetime import date
from sqlalchemy import create_engine, delete, exists, func, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from textual import on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Footer
from textual.worker import Worker, WorkerState
from lazyfeed.db import init_db
from lazyfeed.decorators import fetch_guard, rollback_session
from lazyfeed.feeds import fetch_content, fetch_entries, fetch_feed
from lazyfeed.http_client import http_client_session
from lazyfeed.models import Feed, Item
from lazyfeed.settings import APP_NAME, DB_URL, Settings
from lazyfeed.widgets import (
    CustomHeader,
    ItemTable,
    RSSFeedTree,
    ItemScreen,
)
from lazyfeed.widgets.modals import (
    AddFeedModal,
    EditFeedModal,
    ConfirmActionModal,
    HelpModal,
)
import lazyfeed.messages as messages


class LazyFeedApp(App):
    """
    A simple and modern RSS feed reader for the terminal.
    """

    TITLE = APP_NAME
    ENABLE_COMMAND_PALETTE = False
    CSS_PATH = "global.tcss"

    BINDINGS = [
        Binding("ctrl+c,escape,q", "quit", "quit"),
        Binding("?,f1", "help", "help"),
        Binding("R", "refresh", "refresh"),
    ]

    is_fetching: var[bool] = var(False)
    show_read: var[bool] = var(False)

    def __init__(self, settings: Settings):
        super().__init__()

        self.settings = settings
        self.theme = self.settings.theme

        sort_column = getattr(Item, self.settings.sort_by, Item.published_at)
        self.sort_order = sort_column.desc()
        if self.settings.sort_order == "ascending":
            self.sort_order = sort_column.asc()

        engine = create_engine(f"sqlite:///{DB_URL}")
        init_db(engine)

        Session = sessionmaker(bind=engine)
        self.session = Session()

    def compose(self) -> ComposeResult:
        yield CustomHeader(
            title=f"↪ {APP_NAME}",
            subtitle=f"v{self.settings.version}",
        )
        yield RSSFeedTree(label="*")
        yield ItemTable()
        yield Footer()

    async def on_mount(self) -> None:
        self.item_table = self.query_one(ItemTable)
        self.rss_feed_tree = self.query_one(RSSFeedTree)

        self.item_table.focus()

        await self.sync_feeds()
        await self.sync_items()

        if self.settings.auto_load:
            self.fetch_items()

    def action_help(self) -> None:
        widget = self.focused
        if not widget:
            self.notify("first you have to focus a widget", severity="warning")
            return

        self.push_screen(HelpModal(widget=widget))

    @fetch_guard
    @rollback_session()
    async def action_quit(self) -> None:
        if self.settings.auto_read:
            stmt = update(Item).where(Item.is_read.is_(False)).values(is_read=True)
            self.session.execute(stmt)
            self.session.commit()

        self.session.close()
        self.exit(return_code=0)

    @fetch_guard
    async def action_refresh(self) -> None:
        self.fetch_items()

    def toggle_widget_loading(self, widget: Widget, loading: bool = False) -> None:
        widget.loading = loading

    @on(messages.AddFeed)
    @fetch_guard
    @rollback_session("something went wrong while saving new feed")
    async def add_feed(self) -> None:
        async def callback(response: dict | None = None) -> None:
            if not response:
                return

            title = response.get("title", "")
            url = response.get("url")
            assert url

            async with http_client_session(self.settings) as client_session:
                stmt = select(exists().where(Feed.url == url))
                feed_in_db = self.session.execute(stmt).scalar()
                if feed_in_db:
                    self.notify("feed already exists", severity="error")
                    return

                feed = await fetch_feed(client_session, url, title)
                self.session.add(feed)
                self.session.commit()

                self.notify("new feed added")
                await self.sync_feeds()

        self.push_screen(AddFeedModal(), callback)

    @on(messages.EditFeed)
    @fetch_guard
    @rollback_session("something went wrong while updating feed")
    async def update_feed(self, message: messages.EditFeed) -> None:
        stmt = select(Feed).where(Feed.id == message.id)
        feed_in_db = self.session.execute(stmt).scalar()
        if not feed_in_db:
            self.notify("feed not found", severity="error")
            return

        async def callback(response: dict | None = None) -> None:
            if not response:
                return

            title = response.get("title", "")
            url = response.get("url")
            assert url
            if not title:
                async with http_client_session(self.settings) as client_session:
                    feed = await fetch_feed(client_session, url, title)
                    title = feed.title

            feed_in_db.title = title
            feed_in_db.url = url
            self.session.commit()

            self.notify("feed updated")

            await self.sync_feeds()

        self.push_screen(EditFeedModal(feed_in_db.url, feed_in_db.title), callback)

    @on(messages.DeleteFeed)
    @fetch_guard
    @rollback_session("something went wrong while removing feed")
    async def delete_feed(self, message: messages.DeleteFeed) -> None:
        stmt = select(Feed).where(Feed.id == message.id)
        feed_in_db = self.session.execute(stmt).scalar()
        if not feed_in_db:
            self.notify("feed not found", severity="error")
            return

        async def callback(response: bool | None = False) -> None:
            if not response:
                return

            stmt = delete(Feed).where(Feed.id == feed_in_db.id)
            self.session.execute(stmt)
            self.session.commit()

            self.notify(f'feed "{feed_in_db.title}" removed')

            await self.sync_feeds()
            await self.sync_items()

        self.push_screen(
            ConfirmActionModal(
                border_title="remove feed",
                message=f'are you sure you want to remove "{feed_in_db.title}"?',
                action_name="remove",
            ),
            callback,
        )

    @on(messages.FilterByFeed)
    @fetch_guard
    @rollback_session("something went wrong while getting items from feed")
    async def filter_by_feed(self, message: messages.FilterByFeed) -> None:
        self.show_read = True

        stmt = (
            select(Item).where(Item.feed_id.is_(message.id)).order_by(self.sort_order)
        )
        results = self.session.execute(stmt).scalars().all()

        self.item_table.mount_items(results)
        self.item_table.focus()

    @on(messages.MarkAsRead)
    @rollback_session("something went wrong while updating item")
    async def mark_item_as_read(self, message: messages.MarkAsRead) -> None:
        item_id = message.item_id

        stmt = select(Item).where(Item.id == item_id)
        result = self.session.execute(stmt).scalar()
        if not result:
            self.notify("something went wrong while getting the item", severity="error")
            return

        stmt = update(Item).where(Item.id == item_id).values(is_read=not result.is_read)
        self.session.execute(stmt)
        self.session.commit()

        self.session.refresh(result)

        if self.show_read:
            self.item_table.update_item(f"{item_id}", result)
        else:
            self.item_table.remove_row(row_key=f"{item_id}")

        self.item_table.border_subtitle = f"{self.item_table.row_count}"

        stmt = select(Feed).where(Feed.id == result.feed_id)
        result = self.session.execute(stmt).scalar()
        if result:
            stmt = select(
                func.coalesce(func.count(Item.id).filter(Item.is_read.is_(False)), 0)
            ).where(Item.feed_id == result.id)
            pending_posts = self.session.execute(stmt).scalar()

            self.rss_feed_tree.update_feed((result.id, pending_posts, result.title))

    @on(messages.MarkAllAsRead)
    @fetch_guard
    @rollback_session("something went wrong while updating items")
    async def mark_all_items_as_read(self) -> None:
        async def callback(response: bool | None = False) -> None:
            if not response:
                return

            stmt = update(Item).where(Item.is_read.is_(False)).values(is_read=True)
            self.session.execute(stmt)
            self.session.commit()
            self.notify("all items marked as read")

            await self.sync_feeds()
            await self.sync_items()

        if self.settings.confirm_before_read:
            self.push_screen(
                ConfirmActionModal(
                    border_title="mark all as read",
                    message="are you sure that you want to mark all items as read?",
                    action_name="confirm",
                ),
                callback,
            )
        else:
            await callback(True)

    @on(messages.ShowAll)
    @fetch_guard
    @rollback_session(
        error_message="something went wrong while getting items",
        callback=lambda self: self.toggle_widget_loading(self.item_table),
    )
    async def show_all_items(self) -> None:
        self.show_read = True

        stmt = select(Item).order_by(self.sort_order)
        results = self.session.execute(stmt).scalars().all()
        self.item_table.mount_items(results)

    @on(messages.ShowPending)
    @fetch_guard
    async def show_pending_items(self) -> None:
        self.show_read = False
        await self.sync_items()

    @on(messages.Open)
    @fetch_guard
    @rollback_session("something went wrong while updating item")
    async def open_item(self, message: messages.Open) -> None:
        item_id = message.item_id

        stmt = select(Item).where(Item.id == item_id)
        result = self.session.execute(stmt).scalar()
        if result:
            self.push_screen(ItemScreen(result))
            self.post_message(messages.MarkAsRead(item_id))

    @on(messages.OpenInBrowser)
    @fetch_guard
    @rollback_session("something went wrong while updating item")
    async def open_in_browser(self, message: messages.OpenInBrowser) -> None:
        item_id = message.item_id

        stmt = select(Item).where(Item.id == item_id)
        result = self.session.execute(stmt).scalar()
        if result:
            self.open_url(result.url)
            self.post_message(messages.MarkAsRead(item_id))
        else:
            self.notify("item not found", severity="error")

    @on(messages.SaveForLater)
    @fetch_guard
    @rollback_session("something went wrong while updating items")
    async def save_for_later(self, message: messages.SaveForLater) -> None:
        item_id = message.item_id

        stmt = select(Item).where(Item.id == item_id)
        result = self.session.execute(stmt).scalar()
        if result:
            stmt = (
                update(Item)
                .where(Item.id == item_id)
                .values(is_saved=not result.is_saved)
            )
            self.session.execute(stmt)
            self.session.commit()

            self.session.refresh(result)
            self.item_table.update_item(f"{item_id}", result)

    @on(messages.ShowSavedForLater)
    @fetch_guard
    @rollback_session(
        error_message="something went wrong while getting items",
        callback=lambda self: self.toggle_widget_loading(self.item_table),
    )
    async def load_saved_for_later(self) -> None:
        self.show_read = True

        stmt = select(Item).where(Item.is_saved.is_(True)).order_by(self.sort_order)
        results = self.session.execute(stmt).scalars().all()
        self.item_table.mount_items(results)

    @on(messages.ShowToday)
    @fetch_guard
    @rollback_session(
        error_message="something went wrong while getting items",
        callback=lambda self: self.toggle_widget_loading(self.item_table),
    )
    async def load_today_items(self) -> None:
        self.show_read = True

        today = date.today()

        stmt = (
            select(Item)
            .where(func.date(Item.published_at) == today)
            .order_by(self.sort_order)
        )
        results = self.session.execute(stmt).scalars().all()
        self.item_table.mount_items(results)

    @rollback_session(
        error_message="something went wrong while getting feeds",
        callback=lambda self: self.toggle_widget_loading(self.rss_feed_tree),
    )
    async def sync_feeds(self) -> None:
        stmt = (
            select(
                Feed.id,
                func.coalesce(
                    func.count(Item.id).filter(Item.is_read.is_(False)), 0
                ).label("pending_posts"),
                Feed.title,
            )
            .outerjoin(Item)
            .group_by(Feed.id, Feed.title)
            .order_by(Feed.title.asc())
        )
        results = self.session.execute(stmt).all()
        self.rss_feed_tree.mount_feeds(results)

    @rollback_session(
        error_message="something went wrong while getting items",
        callback=lambda self: self.toggle_widget_loading(self.item_table),
    )
    async def sync_items(self) -> None:
        stmt = select(Item)
        if not self.show_read:
            stmt = stmt.where(Item.is_read.is_(False))

        stmt = stmt.order_by(self.sort_order)
        results = self.session.execute(stmt).scalars().all()
        self.item_table.mount_items(results)

    @work(exclusive=True)
    async def fetch_items(self) -> None:
        async with http_client_session(self.settings) as client_session:
            feeds = self.session.query(Feed).all()
            n_feeds = len(feeds)

            for i, feed in enumerate(feeds):
                self.item_table.border_title = f"loading... {i + 1}/{n_feeds}"

                tasks = []
                try:
                    entries, etag = await fetch_entries(
                        client_session, feed.url, feed.etag
                    )
                    if not entries:
                        continue

                    feed.etag = etag
                    for entry in entries:
                        stmt = select(Item).where(Item.url == entry.link)
                        result = self.session.execute(stmt).scalar()
                        if result:
                            continue

                        tasks.append(fetch_content(client_session, entry, feed.id))
                except (RuntimeError, Exception) as e:
                    self.notify(
                        f'something went wrong when parsing feed "{feed.title}": {e}'
                    )

                results = await asyncio.gather(*tasks, return_exceptions=True)
                successful_items = [
                    result for result in results if not isinstance(result, Exception)
                ]
                unique_items = {item.url: item for item in successful_items}
                try:
                    self.session.add_all(list(unique_items.values()))
                    self.session.commit()
                except (IntegrityError, Exception) as e:
                    self.session.rollback()
                    self.notify(f"something went wrong while saving items: {e}")

    @on(Worker.StateChanged)
    async def on_fetch_items_state(self, event: Worker.StateChanged) -> None:
        if event.state == WorkerState.PENDING or event.state == WorkerState.RUNNING:
            self.is_fetching = True
            self.toggle_widget_loading(self.item_table, True)
        else:
            self.is_fetching = False
            self.item_table.border_title = "items"

            await self.sync_items()
            await self.sync_feeds()



from textual import on
from textual.app import ComposeResult
from textual.containers import Container, VerticalScroll
from textual.validation import Function
from textual.widgets import Button, Input, Label
from textual.screen import ModalScreen
from lazyfeed.widgets.validators import is_valid_url


class AddFeedModal(ModalScreen[dict | None]):
    """
    Modal screen for adding a new RSS feed.
    """

    BINDINGS = [
        ("escape,q", "dismiss", "dismiss"),
    ]

    def compose(self) -> ComposeResult:
        with VerticalScroll(classes="modal-body") as container:
            container.border_title = "add new feed"

            yield Container(
                Container(
                    Label("title (optional)"),
                    Input(
                        placeholder="title",
                        classes="input--feed-title",
                    ),
                ),
                Container(
                    Label("url"),
                    Input(
                        placeholder="url",
                        validate_on=["changed", "submitted"],
                        validators=[
                            Function(is_valid_url, "invalid url."),
                        ],
                        classes="input--feed-url",
                    ),
                ),
                classes="inputs",
            )
            yield Button(label="add feed", variant="success", disabled=True)

    def on_mount(self) -> None:
        self.input_fields = self.query(Input)
        self.button = self.query_one(Button)

    def action_dismiss_overlay(self) -> None:
        self.dismiss(None)

    @on(Input.Changed, ".input--feed-url")
    def enable_button(self, event: Input.Changed) -> None:
        if not event.value or not event.validation_result.is_valid:
            self.button.disabled = True
            return

        self.button.disabled = False

    @on(Button.Pressed)
    def submit_form(self) -> None:
        self.dismiss(
            {
                "title": self.query_one(".input--feed-title").value,
                "url": self.query_one(".input--feed-url").value,
            }
        )

    @on(Input.Submitted, ".input--feed-url")
    async def add_new_feed(self, event: Input.Submitted) -> None:
        if not event.validation_result.is_valid:
            return

        self.dismiss(
            {
                "title": self.query_one(".input--feed-title").value,
                "url": event.value,
            }
        )



import pytest
from lazyfeed.widgets.validators import is_valid_url


@pytest.mark.parametrize(
    "url, expected",
    [
        ("", False),
        ("://example.com", False),
        ("ftp://example.com", True),
        ("htp://example.com", False),
        ("http//example.com", False),
        ("http://", False),
        ("http://-example.com", False),
        ("http://.com", False),
        ("http://123.123.123.123:80/resource", True),
        ("http://192.168.1.1", True),
        ("http://example-.com", False),
        ("http://example..com", False),
        ("http://example.co.uk", True),
        ("http://example.com", True),
        ("http://example.com#fragment", False),
        ("http://example.com:8080", True),
        ("http://example.com:abc", False),
        ("http://example.com?query=param", True),
        ("http://localhost", True),
        ("http:/example.com", False),
        ("https://example.com", True),
        ("https://example.com/path/to/resource", True),
        ("https://subdomain.example.com", True),
    ],
)
def test_is_valid_url(url, expected):
    assert is_valid_url(url) == expected



from typing import Tuple, List
from rich.text import Text
from textual.binding import Binding
from textual.widgets import Tree
from lazyfeed.messages import AddFeed, DeleteFeed, EditFeed, FilterByFeed
from lazyfeed.widgets.helpable import HelpData


class RSSFeedTree(Tree):
    """
    Custom Tree that provides functionality to show, add, edit, delete and filter
    by feeds.
    """

    help = HelpData(
        title="rss feed tree",
        description="""\
A tree for managing RSS feeds. You can (`a`)dd new feeds, (`e`)dit an existing feed, 
(`d`)elete feeds, or select one to filter the items (`enter`).
""",
    )

    BINDINGS = [
        Binding("backspace,d,x", "delete", "delete feed"),
        Binding("a,n", "add", "add"),
        Binding("e", "edit", "edit"),
        Binding("enter", "select_feed", "select"),
        Binding("up,k", "cursor_up", "cursor up", show=False),
        Binding("down,j", "cursor_down", "cursor down", show=False),
        Binding("g", "scroll_home", "cursor to top", show=False),
        Binding("G", "scroll_end", "cursor to bottom", show=False),
    ]

    def on_mount(self) -> None:
        self.show_root = False
        self.border_title = "feeds"

        css_variables = self.app.get_css_variables()
        self.primary = css_variables.get("primary", "blue")

    def action_delete(self) -> None:
        if not self.cursor_node or not self.cursor_node.data:
            self.notify("no feed selected")
            return

        self.post_message(DeleteFeed(self.cursor_node.data["id"]))

    def action_add(self) -> None:
        self.post_message(AddFeed())

    def action_edit(self) -> None:
        if not self.cursor_node or not self.cursor_node.data:
            self.notify("no feed selected")
            return

        self.post_message(EditFeed(id=self.cursor_node.data["id"]))

    def action_select_feed(self) -> None:
        if not self.cursor_node or not self.cursor_node.data:
            self.notify("no feed selected")
            return

        self.post_message(FilterByFeed(id=self.cursor_node.data["id"]))

    def update_feed(self, feed_data: Tuple[int, int, str]) -> None:
        feed_id, pending_posts, title = feed_data
        for node in self.root.children:
            if node.data and node.data["id"] == feed_id:
                label = Text()
                if pending_posts:
                    label.append(f"{pending_posts}", style=f"on {self.primary}")
                    label.append(" ")

                label.append(title)

                node.label = label
                self.refresh()
                return

    def mount_feeds(self, feeds_data: List[Tuple[int, int, str]]) -> None:
        self.clear()
        self.guide_depth = 3
        self.root.expand()

        for feed in feeds_data:
            feed_id, pending_posts, title = feed
            label = Text()
            if pending_posts:
                label.append(f"{pending_posts}", style=f"on {self.primary}")
                label.append(" ")

            label.append(title)

            self.root.add_leaf(label=label, data={"id": feed_id})

        self.cursor_line = 0



from textual.binding import Binding
from textual.widgets import DataTable
from textual.widgets.data_table import CellDoesNotExist
from lazyfeed.models import Item
from lazyfeed.messages import (
    MarkAllAsRead,
    MarkAsRead,
    Open,
    OpenInBrowser,
    SaveForLater,
    ShowPending,
    ShowAll,
    ShowSavedForLater,
    ShowToday,
)
from lazyfeed.widgets.helpable import HelpData


class ItemTable(DataTable):
    """
    Custom DataTable widget for displaying and managing a list of items.
    """

    help = HelpData(
        title="item table",
        description="""\
Table for managing items from your RSS feeds. You can (`o`)pen an item to read
it in markdown, (`O`)pen it in your default browser, (`m`)ark it as read or (`s`)ave it
for later.
""",
    )

    BINDINGS = [
        Binding("up,k", "cursor_up", "cursor up", show=False),
        Binding("down,j", "cursor_down", "cursor down", show=False),
        Binding("g", "scroll_top", "cursor to top", show=False),
        Binding("G", "scroll_bottom", "cursor to bottom", show=False),
        Binding("o", "open", "open"),
        Binding("O", "open_in_browser", "open in browser"),
        Binding("m", "mark_as_read", "mark as read"),
        Binding("M", "mark_all_as_read", "mark all as read", show=False),
        Binding("s", "save_for_later", "save"),
        Binding("a", "show_pending", "pending"),
        Binding("A", "show_all", "all", show=False),
        Binding("l", "show_saved", "saved"),
        Binding("t", "show_today", "today", show=False),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(cursor_type="row", header_height=0, *args, **kwargs)

    def on_mount(self) -> None:
        self.add_column("items", key="items")
        self.border_title = "items"

    def action_mark_as_read(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        except CellDoesNotExist:
            return

        assert row_key.value
        self.post_message(MarkAsRead(int(row_key.value)))

    def action_mark_all_as_read(self) -> None:
        self.post_message(MarkAllAsRead())

    def action_open(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        except CellDoesNotExist:
            return

        assert row_key.value
        self.post_message(Open(int(row_key.value)))

    def action_open_in_browser(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        except CellDoesNotExist:
            return

        assert row_key.value
        self.post_message(OpenInBrowser(int(row_key.value)))

    def action_save_for_later(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
        except CellDoesNotExist:
            return

        assert row_key.value
        self.post_message(SaveForLater(int(row_key.value)))

    def action_show_all(self) -> None:
        self.post_message(ShowAll())

    def action_show_pending(self) -> None:
        self.post_message(ShowPending())

    def action_show_saved(self) -> None:
        self.post_message(ShowSavedForLater())

    def action_show_today(self) -> None:
        self.post_message(ShowToday())

    def format_item(self, item: Item) -> str:
        saved = "" if item.is_saved else " "
        item_title = (
            f"[bold]{item.title}[/]"
            if not item.is_read
            else f"[bold strike]{item.title}[/]"
        )
        url = item.url

        return f"{saved} [bold]{item_title}[/] ([underline italic]{url}[/])"

    def update_item(self, row_key: str, item: Item) -> None:
        self.update_cell(row_key, "items", self.format_item(item))

    def mount_items(self, items: list[Item]) -> None:
        self.clear()
        for item in items:
            self.add_row(self.format_item(item), key=f"{item.id}")

        self.border_subtitle = f"{self.row_count}"
        self.refresh()



from dataclasses import dataclass, field
from typing import Protocol, runtime_checkable


@dataclass
class HelpData:
    """
    Data related to a widget to be displayed in the help modal.
    """

    # Title of the widget.
    title: str = field(default="")

    # Description in markdown format.
    description: str = field(default="")


@runtime_checkable
class Helpable(Protocol):
    """
    Protocol for widgets that contain information required by the
    help modal.
    """

    help: HelpData



from typing import Callable
from functools import wraps


def rollback_session(
    error_message: str = "",
    severity: str = "error",
    callback: Callable | None = None,
):
    def decorator(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            try:
                return await func(self, *args, **kwargs)
            except Exception as e:
                self.session.rollback()
                message = (
                    f"{error_message}: {e}"
                    if error_message
                    else f"something went wrong: {e}"
                )
                self.notify(message, severity=severity)
            finally:
                if callback:
                    callback(self)

        return wrapper

    return decorator



from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Label, Static


class CustomHeader(Static):
    """
    Custom header widget displaying a title and subtitle.
    """

    def __init__(self, title: str, subtitle: str) -> None:
        self.title = title
        self.subtitle = subtitle
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Label(self.title, classes="header__title"),
            Label(self.subtitle, classes="header__subtitle"),
        )
